#include "goldworker.hpp"

GOLDWorker::GOLDWorker(InputParameter input, QObject *parent) : QObject(parent)
{
    /* Input variables parameters */
    is_desktop = input.isDesktop;
    is_simulation = input.isSimulation;

    is_input_real_time = !input.isFileInput;
    save_output = input.isFileOutput;
    in_file_name = input.inputFileName;
    out_file_name = input.outputFileName;
    in_gstream = input.inputGStreamer;

    scale = input.scale;
    target_fps = input.targetFps;

    m = input.m;
    threshold = input.threshold;
    min_val = input.minVal;
    m_filter = input.mFilter;

    polynomium_degree = input.polynomium_degree;
    center_points = input.center_points;

    geometric_output = (float*) malloc(sizeof (float) * static_cast<unsigned int>(polynomium_degree));

    height = input.height;
    pitch = input.pitch;
    dist_ahead = input.distAhead;
    space_to_one_side = input.spaceToOneSide;
    bottom_offset = input.bottomOffset;

    input_delay = input.inputDelay;
    median_filter_size = input.medianFilterSize;
    temporal_filter_offset = input.temporalFilterOffset;
    temporal_filter_threshold = input.temporalFilterThreshold;

    /* Fixed parameters */
    bev_image_size[0] = static_cast<float>(H)/scale;
    bev_image_size[1] = static_cast<float>(W)/scale;

    principal_point[0] = static_cast<float>(W)/2;
    principal_point[1] = static_cast<float>(H)/2;

    end_of_stream = false;
    end_of_pre_processing = false;
    end_of_processing = false;
    end_of_post_processing = false;
    is_output_on_file = false;
    start_flag = false;

    last_index = -2;
    last_index_counter = 0;

    log_file = fopen("/home/root/GOLDLaneDetector/log/log.txt", "a");
}

GOLDWorker::~GOLDWorker()
{
    fclose(log_file);
    sem_close(semaphore);
    sem_unlink(SHSEM);
    munmap(shm_ptr, shm_size);
    close(shm_fd);
    shm_unlink(SHMEM);
}

void GOLDWorker::process()
{
    std::cout.precision(5);

    if(pthread_mutex_init(&lock, nullptr) != 0)
    {
        std::cout << "Lock initialization has failed."
                  << std::endl;
        emit error("");
        return;
    }

    std::cout << "Opening video stream" << std::endl;

    cv::VideoCapture capture;
    get_capture(capture);    

    std::cout << "Checking video stream" << std::endl;

    if(is_desktop)
    {
        if(!capture.isOpened())
        {
            std::cout << "Video stream not opened or file not present."
                      << std::endl;
            emit error("");
            return;
        }
    }
    else
    {
        if(!capture.isOpened())
        {
            std::cout << "Video stream not opened or file not present under /home/root/GOLDLaneDetector/input."
                      << std::endl;
            emit error("");
            return;
        }
    }

    std::cout << "Video stream opened" << std::endl;
    std::cout << "Video lenght: " << capture.get(cv::CAP_PROP_FRAME_COUNT) << std::endl;

    pthread_mutex_lock(&lock);

    if(in_file_name != "")
        fprintf(log_file, "File name: %s\n", in_file_name.toStdString().c_str());
    else
        fprintf(log_file, "Stream name: %s\n", in_gstream.toStdString().c_str());

    pthread_mutex_unlock(&lock);

    if(save_output)
    {
        if(is_desktop)
        {
            video = new cv::VideoWriter(out_file_name.toStdString(), cv::VideoWriter::fourcc('M','J','P','G'),
                                        target_fps, cv::Size(W, H), true);
            if(!video->isOpened())
            {
                std::cout << "The video named \"" << out_file_name.toStdString() << "\" cannot be opened." << std::endl;
                emit error("");
                return;
            }
        }
        else
        {
            video = new cv::VideoWriter("/home/root/GOLDLaneDetector/output/" + out_file_name.toStdString(),
                                        cv::VideoWriter::fourcc('M','J','P','G'), target_fps, cv::Size(W, H), true);
            if(!video->isOpened())
            {
                std::cout << "The video named \"" << out_file_name.toStdString() << "\" under /home/root/GOLDLaneDetector/output"
                          << "cannot be opened." << std::endl;
                emit error("");
                return;
            }
        }
        is_output_on_file = true;
    }

    generate_transformation_matrix(focal_length, principal_point, image_size, height, pitch, dist_ahead,
            space_to_one_side, bottom_offset, out_width, transformation_matrix, image_to_vehicle_matrix, out_size);

    generate_transformation_matrix_inverse(focal_length, principal_point, image_size, height, pitch, dist_ahead,
            space_to_one_side, bottom_offset, out_width, transformation_matrix_inv, out_size);

    int o_flag = O_RDWR | O_CREAT;
    int mmap_flag;

    shm_fd = shm_open(SHMEM, o_flag, 0777);
    if (shm_fd < 0)
    {
        std::cout << "The geometric output shared memory cannot be opened." << std::endl;
        emit error("");
        return;
    }

    mmap_flag = PROT_WRITE;

    shm_size = sizeof(geometric_output_t) + sizeof(float) * static_cast<unsigned int>(polynomium_degree);
    ftruncate(shm_fd, static_cast<long>(shm_size));

    shm_ptr = (geometric_output_t *) mmap(nullptr, shm_size, mmap_flag, MAP_SHARED, shm_fd, 0);
    if (shm_ptr == MAP_FAILED)
    {
        std::cout << "The geometric output shared memory cannot be opened." << std::endl;
        emit error("");
        return;
    }

    semaphore = sem_open(SHSEM, o_flag, 0777, 1);
    if(semaphore == SEM_FAILED)
    {
        std::cout << "The geometric output shared semaphore cannot be opened." << std::endl;
        emit error("");
        return;
    }

    std::cout << "Starting program" << std::endl;

    pthread_t id1 = 0, id2 = 0, id3 = 0, id4 = 0;
    int error_code;
    cpu_set_t cpuset;
    pthread_t thread = pthread_self();

    ThreadLauncher preprocessing_launcher(this, nullptr, PREPROCESSING);
    error_code = pthread_create(&id1, nullptr, launch_thread, &preprocessing_launcher);
    if (error_code != 0)
    {
        fprintf(stderr, "Pre-processing thread can't be created :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }

    CPU_ZERO(&cpuset);
    CPU_SET(1, &cpuset);

    error_code = pthread_setaffinity_np(id1, sizeof(cpu_set_t), &cpuset);
    if (error_code != 0)
    {
        fprintf(stderr, "Cannot set pre-processing thread affinity to CPU 1 :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }
    printf("Pre-processing thread created\n");

    ThreadLauncher processing_launcher(this, nullptr, PROCESSING);
    error_code = pthread_create(&id2, nullptr, launch_thread, &processing_launcher);
    if (error_code != 0)
    {
        fprintf(stderr, "Processing thread can't be created :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }

    CPU_ZERO(&cpuset);
    CPU_SET(2, &cpuset);

    error_code = pthread_setaffinity_np(id2, sizeof(cpu_set_t), &cpuset);
    if (error_code != 0)
    {
        fprintf(stderr, "Cannot set processing thread affinity to CPU 2 :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }
    printf("Processing thread created\n");

    ThreadLauncher postprocessing_launcher(this, nullptr, POSTPROCESSING);
    error_code = pthread_create(&id3, nullptr, launch_thread, &postprocessing_launcher);
    if (error_code != 0)
    {
        fprintf(stderr, "Post-processing thread can't be created :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }

    CPU_ZERO(&cpuset);
    CPU_SET(3, &cpuset);

    error_code = pthread_setaffinity_np(id3, sizeof(cpu_set_t), &cpuset);
    if (error_code != 0)
    {
        fprintf(stderr, "Cannot set processing thread affinity to CPU 3 :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }
    printf("Post-processing thread created\n");

    ThreadLauncher get_inputs_launcher(this, &capture, GET_INPUTS);
    error_code = pthread_create(&id4, nullptr, launch_thread, &get_inputs_launcher);
    if (error_code != 0)
    {
        fprintf(stderr, "Input thread can't be created :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }

    CPU_ZERO(&cpuset);
    CPU_SET(4, &cpuset);

    error_code = pthread_setaffinity_np(id4, sizeof(cpu_set_t), &cpuset);
    if (error_code != 0)
    {
        fprintf(stderr, "Cannot set input thread affinity to CPU 4 :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }
    printf("Input thread created\n");

    CPU_ZERO(&cpuset);
    CPU_SET(5, &cpuset);

    error_code = pthread_setaffinity_np(thread, sizeof(cpu_set_t), &cpuset);
    if (error_code != 0)
    {
        fprintf(stderr, "Cannot set output thread affinity to CPU 0 :[%s]\n", strerror(error_code));
        emit error("");
        return;
    }

    struct sched_param param;
    pthread_setschedparam(pthread_self(), SCHED_RR, &param);
    pthread_setschedparam(id1, SCHED_RR, &param);
    pthread_setschedparam(id2, SCHED_RR, &param);
    pthread_setschedparam(id3, SCHED_RR, &param);
    pthread_setschedparam(id4, SCHED_RR, &param);

    printf("Output thread started\n");

    cv::Mat res, init_frame;
    double min_elapsed_time = 1000, max_elapsed_time = 0;
    double mean_elapsed_time = 0;
    int count = 0, good_count = 0;

    while(!get_start_flag());

    auto start = boost::chrono::system_clock::now();

    while(!get_end_of_post_processing() || !out_images_empty())
    {
        // polling
        while(!out_images_pop_front(res))
        {
            if(get_end_of_post_processing())
            {
                goto end_output;
            }
        }       

        auto out_start = boost::chrono::system_clock::now();

        cv::Mat init_frame;
        // polling
        while(!overlay_images_pop_front(init_frame))
        {
            if(get_end_of_stream())
            {
                goto end_output;
            }
        }

        if(!init_frame.empty())
        {
            res = overlay_frames(init_frame, res);
        }

        if(is_output_on_file)
        {
            video->write(res);
        }

        cv::cvtColor(res, res, cv::COLOR_BGR2RGB);

        emit request_new_content(QPixmap::fromImage(QImage(res.data, res.cols, res.rows,
                                 static_cast<int>(res.step), QImage::Format_RGB888)));
        printf("\rPrinted frame number %d", count);

        auto out_end = boost::chrono::system_clock::now();
        boost::chrono::duration<double> elapsed_seconds_frame = out_end - out_start;

        if(elapsed_seconds_frame.count() > max_elapsed_time)
            max_elapsed_time = elapsed_seconds_frame.count();
        if(elapsed_seconds_frame.count() < min_elapsed_time)
            min_elapsed_time = elapsed_seconds_frame.count();

        count++;
        mean_elapsed_time += elapsed_seconds_frame.count();

        if(elapsed_seconds_frame.count() < (float) 1/target_fps)
        {
            good_count++;
        }
    }

end_output:
    printf("\n\nOutput thread ended\n");
    printf("Output mean elapsed time: %.5f s, max: %.5f s, min: %.5f s\n",
            mean_elapsed_time/count, max_elapsed_time, min_elapsed_time);
    printf("Output good elaboration count rate: %.2f %% \n\n",
            (float)good_count/(float)count*100);

    pthread_mutex_lock(&lock);

    fprintf(log_file, "Output mean: %.5f\n", mean_elapsed_time/count);
    fprintf(log_file, "Output perc: %.2f\n", (float)good_count/(float)count*100);

    pthread_mutex_unlock(&lock);

    pthread_join(id1, nullptr);
    pthread_join(id2, nullptr);
    pthread_join(id3, nullptr);
    pthread_join(id4, nullptr);

    auto end = boost::chrono::system_clock::now();
    boost::chrono::duration<double> elapsed_seconds_total = end - start;

    printf("\n\n-------- computed image at image width: %d, image height: %d --------\n", W, H);
    printf("-        total mean elapsed time: %.5f s\n", elapsed_seconds_total.count()/count);
    printf("-        total frame per second : %.5f\n\n", 1/(elapsed_seconds_total.count()/count));

    pthread_mutex_lock(&lock);

    fprintf(log_file, "Total mean: %.5f\n", elapsed_seconds_total.count()/count);
    fprintf(log_file, "Total FPS: %.5f\n", 1/(elapsed_seconds_total.count()/count));
    fprintf(log_file, "Total frames: %d\n", count);
    fprintf(log_file, "------------------------------------------------------------------------\n\n");

    pthread_mutex_unlock(&lock);

    capture.release();

    if(save_output)
    {
        if(video != nullptr)
        {
            video->release();
            delete video;
        }
    }

    pthread_mutex_destroy(&lock);

    emit finished();
}

void GOLDWorker::get_capture(cv::VideoCapture& capture)
{
    if(is_input_real_time)
    {
        capture.open(in_gstream.toStdString(), cv::CAP_GSTREAMER);
    }
    else
    {
        if(is_desktop)
        {
            capture.open(in_file_name.toStdString(), cv::CAP_FFMPEG);
        }
        else
        {
            capture.open("/home/root/GOLDLaneDetector/input/" + in_file_name.toStdString(), cv::CAP_FFMPEG);
        }
    }
}

auto GOLDWorker::get_inputs(void *arg) -> void*
{
    cv::VideoCapture* capture = static_cast<cv::VideoCapture*>(arg);

    double min_elapsed_time = 1000, max_elapsed_time = 0;
    double mean_elapsed_time = 0;
    int count = 0, good_count = 0;

    printf("Input thread started\n");

    if(!is_simulation)
    {
        set_start_flag(true);
        while (true)
        {
            auto start_frame = boost::chrono::system_clock::now();

            cv::Mat init_frame;
            capture->read(init_frame);

            cv::Mat curr_frame;
            if (init_frame.empty())
            {
                fprintf(stdout, "\nEnd of stream\n\n");
                set_end_of_stream(true);
                break;
            }

            overlay_images_push_back(init_frame);

            cv::cvtColor(init_frame, init_frame, cv::COLOR_BGR2GRAY);
            // cv Mat is row-major, birds_eye_view uses column-major representation
            cv::transpose(init_frame, init_frame);

            in_images_push_back(init_frame);

            auto end_frame = boost::chrono::system_clock::now();

            boost::chrono::duration<double> elapsed_seconds_frame  = end_frame-start_frame;

            if(elapsed_seconds_frame.count() > max_elapsed_time)
                max_elapsed_time = elapsed_seconds_frame.count();
            if(elapsed_seconds_frame.count() < min_elapsed_time)
                min_elapsed_time = elapsed_seconds_frame.count();

            count++;
            mean_elapsed_time += elapsed_seconds_frame.count();

            if(elapsed_seconds_frame.count() < (float)1/target_fps)
            {
                good_count++;
            }

            printf("\rRead frame number %d", count);
        }
    }
    else
    {
        while (true)
        {
            cv::Mat init_frame;
            capture->read(init_frame);

            cv::Mat curr_frame;
            if (init_frame.empty())
            {
                if(!get_start_flag())
                {
                    set_start_flag(true);
                }
                printf("\nStart process\n");
                break;
            }

            if(init_frame.cols != W || init_frame.rows != H)
            {
                cv::resize(init_frame, init_frame, cv::Size(W, H));
            }

            test_images_tmp_push_back(init_frame);

            printf("\rRead frame number %d", count++);
        }

        count = 0;

        while (true)
        {
            auto start_frame = boost::chrono::system_clock::now();

            cv::Mat init_frame;
            bool ret = test_images_tmp_pop_front(init_frame);

            cv::Mat curr_frame;
            if (!ret)
            {
                fprintf(stdout, "\nEnd of stream\n\n");
                set_end_of_stream(true);
                break;
            }

            overlay_images_push_back(init_frame);

            cv::cvtColor(init_frame, init_frame, cv::COLOR_BGR2GRAY);
            // cv Mat is row-major, birds_eye_view uses column-major representation
            cv::transpose(init_frame, init_frame);

            in_images_push_back(init_frame);

            auto tmp_end_frame = boost::chrono::system_clock::now();
            boost::chrono::duration<double> tmp_elapsed_seconds_frame = tmp_end_frame-start_frame;

            double delay = (1/(double)target_fps - tmp_elapsed_seconds_frame.count()) * 1000;
            delay = delay < 1 ? 2 : delay;
            usleep(static_cast<unsigned int>((delay-1)*1000));

            auto end_frame = boost::chrono::system_clock::now();

            boost::chrono::duration<double> elapsed_seconds_frame = end_frame-start_frame;

            if(elapsed_seconds_frame.count() > max_elapsed_time)
                max_elapsed_time = elapsed_seconds_frame.count();
            if(elapsed_seconds_frame.count() < min_elapsed_time)
                min_elapsed_time = elapsed_seconds_frame.count();

            count++;
            mean_elapsed_time += elapsed_seconds_frame.count();

            if(elapsed_seconds_frame.count() < (float) 1/target_fps)
            {
                good_count++;
            }
        }
    }

    printf("\n\nInput thread ended \n");

    printf("Sending mean elapsed time: %.5f s, max: %.5f s, min: %.5f s\n",
            mean_elapsed_time/count, max_elapsed_time, min_elapsed_time);
    printf("Sending good elaboration count rate: %.2f %% \n\n",
            (float)good_count/(float)count*100);

    pthread_mutex_lock(&lock);

    fprintf(log_file, "Input mean: %.5f\n", mean_elapsed_time/count);
    fprintf(log_file, "Input perc: %.2f\n", (float)good_count/(float)count*100);

    pthread_mutex_unlock(&lock);

    return nullptr;
}

auto GOLDWorker::preprocessing(void *arg) -> void*
{
    cv::Mat init_frame;
    CV_UNUSED(arg);

    double min_elapsed_time = 1000, max_elapsed_time = 0;
    double mean_elapsed_time = 0;
    int count = 0, good_count = 0;    

    printf("Pre-processing thread started\n");

    while(!get_end_of_stream() || !in_images_empty())
    {
        // polling
        while(!in_images_pop_front(init_frame))
        {
            if(get_end_of_stream())
            {
                goto end_pre_processing;
            }
        }

        auto start = boost::chrono::system_clock::now();

        if(!init_frame.empty())
        {
            /* Pre-processing */
            emxArray_uint8_T *bev_image;
            emxInitArray_uint8_T(&bev_image, 2);

            birds_eye_view(init_frame.data, transformation_matrix, out_size, bev_image);

            cv::Mat pre_processing_image = cv::Mat(static_cast<int>(out_size[1]),
                    static_cast<int>(out_size[0]), CV_8U);
            pre_processing_image.data = bev_image->data;
            cv::transpose(pre_processing_image, pre_processing_image);

            bev_images_push_back(bev_image);
        }

        auto end = boost::chrono::system_clock::now();
        boost::chrono::duration<double> elapsed_seconds_frame = end - start;

        if(elapsed_seconds_frame.count() > max_elapsed_time)
            max_elapsed_time = elapsed_seconds_frame.count();
        if(elapsed_seconds_frame.count() < min_elapsed_time)
            min_elapsed_time = elapsed_seconds_frame.count();

        count++;
        mean_elapsed_time += elapsed_seconds_frame.count();

        if(elapsed_seconds_frame.count() < (float) 1/target_fps)
        {
            good_count++;
        }
    }

end_pre_processing:
    set_end_of_pre_processing(true);
    printf("\n\nPre-processing thread ended \n");

    printf("Pre-processing mean elapsed time: %.5f s, max: %.5f s, min: %.5f s\n",
            mean_elapsed_time/count, max_elapsed_time, min_elapsed_time);
    printf("Pre-processing good elaboration count rate: %.2f %% \n\n",
            (float)good_count/(float)count*100);

    pthread_mutex_lock(&lock);

    fprintf(log_file, "Pre-processing mean: %.5f\n", mean_elapsed_time/count);
    fprintf(log_file, "Pre-processing perc: %.2f\n", (float)good_count/(float)count*100);

    pthread_mutex_unlock(&lock);

    return nullptr;
}

auto GOLDWorker::processing(void *arg) -> void*
{
    CV_UNUSED(arg);

    double min_elapsed_time = 1000, max_elapsed_time = 0;
    double mean_elapsed_time = 0;
    int count = 0, good_count = 0;

    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    ts.tv_nsec += 10*1000;

    printf("Processing thread started\n");

    sem_timedwait(semaphore, &ts);
    shm_ptr->counter = -1;
    sem_post(semaphore);

    while(!get_end_of_pre_processing() || !bev_images_empty())
    {
        emxArray_uint8_T *curr_frame;

        // polling
        while(!bev_images_pop_front(curr_frame))
        {
            if(get_end_of_pre_processing())
            {
                goto end_processing;
            }
        }

        auto start = boost::chrono::system_clock::now();

        if(curr_frame != nullptr)
        {
            /* Processing */

            // Get a column-major array, the mat should have opposite dimensions and then transpose
            cv::Mat frame = cv::Mat(static_cast<int>(out_size[1]), static_cast<int>(out_size[0]), CV_8U);
            frame.data = curr_frame->data;
            cv::transpose(frame, frame);

            cv::Mat res_frame = frame.clone();
            get_filtered_frame(frame, m, threshold, min_val, m_filter, polynomium_degree, center_points,
                               static_cast<int>(dist_ahead), image_to_vehicle_matrix, res_frame, geometric_output);

            sem_timedwait(semaphore, &ts);
            shm_ptr->counter++;
            memcpy(shm_ptr->coefficients, geometric_output, static_cast<unsigned int>(polynomium_degree));
            sem_post(semaphore);

            // res_frame is now row-major, post-processing needs a column-major matrix
            cv::transpose(res_frame, res_frame);

            emxDestroyArray_uint8_T(curr_frame);
            gold_images_push_back(res_frame);
        }

        auto end = boost::chrono::system_clock::now();
        boost::chrono::duration<double> elapsed_seconds_frame = end - start;

        if(elapsed_seconds_frame.count() > max_elapsed_time)
            max_elapsed_time = elapsed_seconds_frame.count();
        if(elapsed_seconds_frame.count() < min_elapsed_time)
            min_elapsed_time = elapsed_seconds_frame.count();

        count++;
        mean_elapsed_time += elapsed_seconds_frame.count();

        if(elapsed_seconds_frame.count() < (float) 1/target_fps)
        {
            good_count++;
        }
    }

end_processing:
    set_end_of_processing(true);
    printf("\n\nProcessing thread ended\n");

    printf("Processing mean elapsed time: %.5f s, max: %.5f s, min: %.5f s\n",
            mean_elapsed_time/count, max_elapsed_time, min_elapsed_time);
    printf("Processing good elaboration count rate: %.2f %% \n\n",
            (float)good_count/(float)count*100);

    pthread_mutex_lock(&lock);

    fprintf(log_file, "Processing mean: %.5f\n", mean_elapsed_time/count);
    fprintf(log_file, "Processing perc: %.2f\n", (float)good_count/(float)count*100);

    pthread_mutex_unlock(&lock);

    return nullptr;
}

auto GOLDWorker::postprocessing(void *arg) -> void*
{   
    CV_UNUSED(arg);

    double min_elapsed_time = 1000, max_elapsed_time = 0;
    double mean_elapsed_time = 0;
    int count = 0, good_count = 0;

    printf("Post-processing thread started\n");

    while(!get_end_of_processing() || !gold_images_empty())
    {
        cv::Mat curr_frame = cv::Mat::zeros(static_cast<int>(out_size[0]),
                static_cast<int>(out_size[1]), CV_8U);

        // polling
        while(!gold_images_pop_front(curr_frame))
        {
            if(get_end_of_processing())
            {
                goto end_post_processing;
            }
        }

        auto start = boost::chrono::system_clock::now();

        if(!curr_frame.empty())
        {
            emxArray_uint8_T *bev_inverse_image;
            emxInitArray_uint8_T(&bev_inverse_image, 2);

            // We obtain a column-major matrix, H and W must be switched
            cv::Mat res_frame = cv::Mat(static_cast<int>(bev_image_size[1]),
                    static_cast<int>(bev_image_size[0]), CV_8U);

            /* Post-processing */
            birds_eye_view_inv(curr_frame.data, transformation_matrix_inv, scale,
                               bev_image_size, bev_inverse_image);

            res_frame.data = bev_inverse_image->data;
            cv::transpose(res_frame, res_frame);

            emxDestroyArray_uint8_T(bev_inverse_image);

            cv::resize(res_frame, res_frame, cv::Size(W, H));
            out_images_push_back(res_frame);
        }

        auto end = boost::chrono::system_clock::now();
        boost::chrono::duration<double> elapsed_seconds_frame = end - start;

        if(elapsed_seconds_frame.count() > max_elapsed_time)
            max_elapsed_time = elapsed_seconds_frame.count();
        if(elapsed_seconds_frame.count() < min_elapsed_time)
            min_elapsed_time = elapsed_seconds_frame.count();

        count++;
        mean_elapsed_time += elapsed_seconds_frame.count();

        if(elapsed_seconds_frame.count() < (float) 1/target_fps)
        {
            good_count++;
        }
    }

end_post_processing:
    set_end_of_post_processing(true);
    printf("\n\nPost-processing thread ended\n");

    printf("Post-processing mean elapsed time: %.5f s, max: %.5f s, min: %.5f s\n",
            mean_elapsed_time/count, max_elapsed_time, min_elapsed_time);
    printf("Post-processing good elaboration count rate: %.2f %% \n\n",
            (float)good_count/(float)count*100);

    pthread_mutex_lock(&lock);

    fprintf(log_file, "Post-processing mean: %.5f\n", mean_elapsed_time/count);
    fprintf(log_file, "Post-processing perc: %.2f\n", (float)good_count/(float)count*100);

    pthread_mutex_unlock(&lock);

    return nullptr;
}

void GOLDWorker::get_filtered_frame(cv::Mat frame, float mF, float thresholdF, float min_val, float m_filterF, int n,
                                    int n_points, int dist_ahead, float transformation[9], cv::Mat& res_frame, float *p)
{
    cv::Size size = frame.size();
    cv::Mat r_frame(size, CV_32F, cv::Scalar(0));
    cv::Mat tmp_res_frame(size, CV_8U, cv::Scalar(0));
    res_frame = cv::Mat::zeros(size, CV_8U);

    int m = static_cast<int>(mF);
    int threshold = static_cast<int>(thresholdF);
    int m_filter = static_cast<int>(m_filterF);

    int jmm, jmp;
    for(int i = 0; i < size.height; i++)
    {
        for(int j = 0; j < size.width; j++)
        {
            if(j - m > 0)
            {
                jmm = j - m;
            }
            else
            {
                jmm = 0;
            }

            if(j + m < size.width)
            {
                jmp = j + m;
            }
            else
            {
                jmp = size.width - 1;
            }

            uint8_t element = frame.at<uint8_t>(i, j);
            uint8_t element_l = frame.at<uint8_t>(i, jmm);
            uint8_t element_r = frame.at<uint8_t>(i, jmp);
            int d1 = static_cast<int>(element) - static_cast<int>(element_l);
            int d2 = static_cast<int>(element) - static_cast<int>(element_r);

            if(d1 > threshold && d2 > threshold)
            {
                r_frame.at<float>(i, j) = d1 + d2;
            }
        }
    }

    //Filtering salt and pepper noise
    cv::medianBlur(r_frame, r_frame, median_filter_size);

    for (int i = 0; i < size.height; i++)
    {
        for(int j = 0; j < size.width; j++)
        {
            if (r_frame.at<float>(i, j) < min_val)
                r_frame.at<float>(i, j) = 0;
        }
    }

    int imm, imp;
    for (int i = 0; i < size.height; i++)
    {
       for(int j = 0; j < size.width; j++)
       {
           if ((i - m_filter) > 0)
               imm = i - m_filter;
           else
               imm = 0;

           if ((i + m_filter) < size.height)
               imp = i + m_filter;
           else
               imp = size.height - 1;

           if(static_cast<int>(r_frame.at<float>(i, j)) > 1)
           {
               if((static_cast<int>(r_frame.at<float>(imp, j)) > 1 ||
                   static_cast<int>(r_frame.at<float>(imm, j)) > 1))
               {
                   tmp_res_frame.at<unsigned char>(i, j) = 255U;
               }
               else
               {
                   tmp_res_frame.at<unsigned char>(i, j) = 0U;
               }
           }
           else
           {
               tmp_res_frame.at<unsigned char>(i, j) = 0U;
           }
       }
    }

    std::vector<cv::Point> valid_points;

    int *values = (int*) malloc(static_cast<unsigned long>(size.width) * sizeof (int));
    for(int i = 0; i < size.width; i++)
    {
       values[i] = 0;
    }

    for(int j = 0; j < size.width; j++)
    {
       for(int i = 0; i < size.height; i++)
       {
           if(tmp_res_frame.at<unsigned char>(i, j) != 0U)
           {
               values[j]++;
           }
       }
    }

    int max_line = 0;
    int max_index_1 = -1;
    for(int i = static_cast<int>(size.width/2); i < size.width; i++)
    {
       if(values[i] > max_line)
       {
           max_line = values[i];
           max_index_1 = i;
       }
    }

    bool flag = false;
    if(max_index_1 != -1)
    {
       for(int j = max_index_1 + 1; j < size.width; j++)
       {
           if(!flag && values[j] == 0)
           {
               flag = true;
           }

           if(flag)
           {
               for(int i = 0; i < size.height; i++)
               {
                   tmp_res_frame.at<unsigned char>(i, j) = 0U;
               }
           }
       }
    }

    max_line = 0;
    int max_index_2 = -1;
    for(int i = static_cast<int>(size.width/2); i > 0; i--)
    {
       if(values[i] > max_line)
       {
           max_line = values[i];
           max_index_2 = i;
       }
    }

    flag = false;
    if(max_index_2 != -1)
    {
       for(int j = max_index_2 - 1; j > 0; j--)
       {
           if(!flag && values[j] == 0)
           {
               flag = true;
           }

           if(flag)
           {
               for(int i = 0; i < size.height; i++)
               {
                   tmp_res_frame.at<unsigned char>(i, j) = 0U;
               }
           }
       }
    }

    if(max_index_1 != -1 && max_index_2 != -1)
    {
       flag = false;
       for(int j = max_index_2; j < size.width/2; j++)
       {
           if(!flag && values[j] == 0)
           {
               flag = true;
           }

           if(flag)
           {
               for(int i = 0; i < size.height; i++)
               {
                   tmp_res_frame.at<unsigned char>(i, j) = 0U;
               }
           }
       }

       flag = false;
       for(int j = max_index_1; j > size.width/2; j--)
       {
           if(!flag && values[j] == 0)
           {
               flag = true;
           }

           if(flag)
           {
               for(int i = 0; i < size.height; i++)
               {
                   tmp_res_frame.at<unsigned char>(i, j) = 0U;
               }
           }
       }
    }

    free(values);

    for(int i = 0; i < size.height; i++)
    {
       int left_index = -1;
       for (int j = size.width/2; j >= 0; j--)
       {
           if(tmp_res_frame.at<unsigned char>(i, j) != 0U)
           {
               left_index = j;
               break;
           }
       }

       int right_index = -1;
       for (int j = size.width/2; j < size.width; j++)
       {
           if(tmp_res_frame.at<unsigned char>(i, j) != 0U)
           {
               right_index = j;
               break;
           }
       }

       if(right_index == -1 || left_index == -1)
       {
           continue;
       }

       valid_points.emplace_back(static_cast<int>((right_index + left_index)/2), i);
    }

    res_frame = tmp_res_frame.clone();

    if(valid_points.size() > 3)
    {
       // Lane center
       emxArray_real32_T *image_points;
       image_points = emxCreate_real32_T(static_cast<int>(valid_points.size()), 2);

       emxArray_real32_T *p_real = nullptr;
       p_real = emxCreate_real32_T(n, 1);

       unsigned int size_0 = static_cast<unsigned int>(image_points->size[0U]);

       for (unsigned int idx0 = 0; idx0 < size_0; idx0++)
       {
           image_points->data[idx0] = (float) valid_points[idx0].x;
           image_points->data[idx0 + size_0] = (float) valid_points[idx0].y;
       }

       cv::transpose(res_frame, res_frame);
       image_to_vehicle(res_frame.data, transformation, image_points, n, n_points, dist_ahead, p_real);
       memcpy(p, p_real->data, static_cast<unsigned long>(n));
       cv::transpose(res_frame, res_frame);

       emxDestroyArray_real32_T(image_points);
       emxDestroyArray_real32_T(p_real);
    }
}

void GOLDWorker::set_start_flag(bool value)
{
    pthread_mutex_lock(&lock);
    start_flag = value;
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::get_start_flag() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = start_flag;
    pthread_mutex_unlock(&lock);

    return res;
}

void GOLDWorker::set_end_of_stream(bool value)
{
    pthread_mutex_lock(&lock);
    end_of_stream = value;
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::get_end_of_stream() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = end_of_stream;
    pthread_mutex_unlock(&lock);

    return res;
}

void GOLDWorker::set_end_of_pre_processing(bool value)
{
    pthread_mutex_lock(&lock);
    end_of_pre_processing = value;
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::get_end_of_pre_processing() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = end_of_pre_processing;
    pthread_mutex_unlock(&lock);

    return res;
}

void GOLDWorker::set_end_of_processing(bool value)
{
    pthread_mutex_lock(&lock);
    end_of_processing = value;
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::get_end_of_processing() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = end_of_processing;
    pthread_mutex_unlock(&lock);

    return res;
}

void GOLDWorker::set_end_of_post_processing(bool value)
{
    pthread_mutex_lock(&lock);
    end_of_post_processing = value;
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::get_end_of_post_processing() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = end_of_post_processing;
    pthread_mutex_unlock(&lock);

    return res;
}

void GOLDWorker::in_images_push_back(const cv::Mat& src)
{
    pthread_mutex_lock(&lock);
    in_images.push_back(src);
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::in_images_pop_front(cv::Mat& res) -> bool
{
    pthread_mutex_lock(&lock);
    if(in_images.empty())
    {
        pthread_mutex_unlock(&lock);
        return false;
    }
    res = in_images.at(0);
    in_images.erase(in_images.begin());
    pthread_mutex_unlock(&lock);

    return true;
}

auto GOLDWorker::in_images_empty() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = in_images.empty();
    pthread_mutex_unlock(&lock);
    return res;
}

void GOLDWorker::bev_images_push_back(emxArray_uint8_T * src)
{
    pthread_mutex_lock(&lock);
    bev_images.push_back(src);
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::bev_images_pop_front(emxArray_uint8_T* &res) -> bool
{
    pthread_mutex_lock(&lock);
    if(bev_images.empty())
    {
        pthread_mutex_unlock(&lock);
        return false;
    }
    res = bev_images.at(0);
    bev_images.erase(bev_images.begin());
    pthread_mutex_unlock(&lock);

    return true;
}

auto GOLDWorker::bev_images_empty() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = bev_images.empty();
    pthread_mutex_unlock(&lock);
    return res;
}

void GOLDWorker::gold_images_push_back(const cv::Mat& src)
{
    pthread_mutex_lock(&lock);
    gold_images.push_back(src);
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::gold_images_pop_front(cv::Mat& res) -> bool
{
    pthread_mutex_lock(&lock);
    if(gold_images.empty())
    {
        pthread_mutex_unlock(&lock);
        return false;
    }
    res = gold_images.at(0);
    gold_images.erase(gold_images.begin());
    pthread_mutex_unlock(&lock);

    return true;
}

auto GOLDWorker::gold_images_empty() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = gold_images.empty();
    pthread_mutex_unlock(&lock);
    return res;
}

void GOLDWorker::out_images_push_back(const cv::Mat& src)
{
    pthread_mutex_lock(&lock);
    out_images.push_back(src);
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::out_images_pop_front(cv::Mat& res) -> bool
{
    pthread_mutex_lock(&lock);
    if(out_images.empty())
    {
        pthread_mutex_unlock(&lock);
        return false;
    }
    res = out_images.at(0);
    out_images.erase(out_images.begin());
    pthread_mutex_unlock(&lock);

    return true;
}

auto GOLDWorker::out_images_empty() -> bool
{
    bool res;
    pthread_mutex_lock(&lock);
    res = out_images.empty();
    pthread_mutex_unlock(&lock);
    return res;
}

void GOLDWorker::overlay_images_push_back(const cv::Mat& src)
{
    pthread_mutex_lock(&lock);
    overlay_images.push_back(src);
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::overlay_images_pop_front(cv::Mat& res) -> bool
{
    pthread_mutex_lock(&lock);
    if(overlay_images.empty())
    {
        pthread_mutex_unlock(&lock);
        return false;
    }
    res = overlay_images.at(0);
    overlay_images.erase(overlay_images.begin());
    pthread_mutex_unlock(&lock);

    return true;
}

void GOLDWorker::test_images_tmp_push_back(const cv::Mat& src)
{
    pthread_mutex_lock(&lock);
    test_images_tmp.push_back(src);
    pthread_mutex_unlock(&lock);
}

auto GOLDWorker::test_images_tmp_pop_front(cv::Mat& res) -> bool
{
    pthread_mutex_lock(&lock);
    if(test_images_tmp.empty())
    {
        pthread_mutex_unlock(&lock);
        return false;
    }
    res = test_images_tmp.at(0);
    test_images_tmp.erase(test_images_tmp.begin());
    pthread_mutex_unlock(&lock);

    return true;
}

auto GOLDWorker::overlay_frames(cv::Mat base_frame, cv::Mat overlaying_frame) -> cv::Mat
{
    cv::Mat res_frame;

    for (int x = 0; x < overlaying_frame.rows; x++)
    {
        for (int y = 0; y < overlaying_frame.cols; y++)
        {            
            cv::Vec3b color(0, overlaying_frame.at<uchar>(x, y), 0);
            if (color[1] != 0)
            {
                base_frame.at<cv::Vec3b>(x, y) = color;
            }
        }
    }

    res_frame = base_frame;
    return res_frame;
}

