#ifndef GOLDWORKER_H
#define GOLDWORKER_H

#include <QObject>
#include <QDir>
#include <QPixmap>
#include <QString>

#include "opencv2/opencv.hpp"
#include "opencv2/imgproc/imgproc.hpp"

#include <string>
#include <stdarg.h>
#include <iostream>
#include <ctime>
#include <cstdio>
#include <boost/chrono/chrono.hpp>
#include <vector>
#include <map>
#include <pthread.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/shm.h>
#include <semaphore.h>
#include "inputparameter.h"

extern "C"
{
#include "MatlabCode/generate_transformation_matrix/generate_transformation_matrix.h"
#include "MatlabCode/generate_transformation_matrix/generate_transformation_matrix_types.h"
#include "MatlabCode/generate_transformation_matrix_inverse/generate_transformation_matrix_inverse.h"
#include "MatlabCode/birds_eye_view/birds_eye_view.h"
#include "MatlabCode/birds_eye_view_inv/birds_eye_view_inv.h"
#include "MatlabCode/common/birds_eye_view_emxAPI.h"
#include "MatlabCode/common/birds_eye_view_emxutil.h"
#include "MatlabCode/image_to_vehicle/image_to_vehicle.h"
#include "MatlabCode/image_to_vehicle/image_to_vehicle_types.h"
}

#define W 1920
#define H 1080
#define F_W 150
#define SHMEM "GOLDLikeLaneDetectorSharedMemory"
#define SHSEM "/GOLDLikeLaneDetectorSharedMemory"

class GOLDWorker : public QObject
{
    Q_OBJECT
public:
    explicit GOLDWorker(InputParameter input, QObject *parent = nullptr);

    ~GOLDWorker();

public slots:
    void process();

signals:
    void finished();
    void error(QString err);
    void request_new_content(const QPixmap &);

private:

    typedef struct geometric_output
    {
        long counter;
        float coefficients[];
    } geometric_output_t;

    enum ThreadTypes
    {
        PREPROCESSING,
        PROCESSING,
        POSTPROCESSING,
        GET_INPUTS
    };

    class ThreadLauncher
    {

    public:
        ThreadLauncher(GOLDWorker *worker, void *arg, ThreadTypes type)
        {
            this->worker = worker;
            this->arg = arg;
            this->type = type;
        }

        void *launch()
        {
            if(type == PREPROCESSING)
            {
                return worker->preprocessing(arg);
            }
            if(type == PROCESSING)
            {
                return worker->processing(arg);
            }
            if(type == POSTPROCESSING)
            {
                return worker->postprocessing(arg);
            }
            if(type == GET_INPUTS)
            {
                return worker->get_inputs(arg);
            }
            return nullptr;
        }

        GOLDWorker *worker;
        void *arg;
        ThreadTypes type;
    };

    cv::VideoWriter *video;

    std::vector<cv::Mat> in_images;
    std::vector<cv::Mat> overlay_images;
    std::vector<cv::Mat> test_images_tmp;
    std::vector<cv::Mat> out_images;

    std::vector<emxArray_uint8_T *>  bev_images;
    std::vector<cv::Mat> gold_images;

    FILE *log_file;

    geometric_output_t *shm_ptr;
    sem_t *semaphore;
    int shm_fd;
    size_t shm_size;

    pthread_mutex_t lock;
    bool end_of_stream;
    bool end_of_pre_processing;
    bool end_of_processing;
    bool end_of_post_processing;
    bool is_output_on_file;
    bool start_flag;
    bool is_desktop;
    bool is_simulation;

    float focal_length[2] = {static_cast<float>(2015.422939513609), static_cast<float>(2002.217072647680)};
    float principal_point[2];
    float height;
    float pitch;
    float dist_ahead;
    float space_to_one_side;
    float bottom_offset;
    float out_width = F_W;
    float image_size[2] = {H, W};
    float image_to_vehicle_matrix[9];
    float *geometric_output;
    int polynomium_degree;
    int center_points;

    int scale;
    int target_fps;

    int last_index;
    int last_index_counter;

    float transformation_matrix[9];
    float transformation_matrix_inv[9];
    float out_size[2];
    float bev_image_size[2];
    float m, threshold, min_val, m_filter;

    int start_value;

    bool is_input_real_time;
    bool save_output;
    QString in_file_name, out_file_name;
    QString in_gstream;
    int input_delay;
    int median_filter_size;
    int temporal_filter_offset;
    int temporal_filter_threshold;

    static void *launch_thread(void *arg)
    {
        ThreadLauncher* launcher = reinterpret_cast<ThreadLauncher*>(arg);
        return launcher->launch();
    }

    auto get_inputs(void *arg) -> void*;
    void* preprocessing(void *arg);
    auto processing(void *arg) -> void*;
    auto postprocessing(void *arg) -> void*;

    void get_filtered_frame(cv::Mat frame, float mF, float thresholdF, float min_val, float m_filterF, int n,
                            int n_points, int dist_ahead, float transformation[9], cv::Mat& res_frame, float *p);
    auto overlay_frames(cv::Mat base_frame, cv::Mat overlaying_frame) -> cv::Mat;

    void overlay_images_push_back(const cv::Mat& src);
    auto overlay_images_pop_front(cv::Mat& res) -> bool;

    void in_images_push_back(const cv::Mat& src);
    auto in_images_pop_front(cv::Mat& res) -> bool;
    auto in_images_empty() -> bool;

    void test_images_tmp_push_back(const cv::Mat& src);
    auto test_images_tmp_pop_front(cv::Mat& res) -> bool;

    void bev_images_push_back(emxArray_uint8_T * src);
    auto bev_images_pop_front(emxArray_uint8_T* &res) -> bool;
    auto bev_images_empty() -> bool;

    void gold_images_push_back(const cv::Mat& src);
    auto gold_images_pop_front(cv::Mat& res) -> bool;
    auto gold_images_empty() -> bool;

    void out_images_push_back(const cv::Mat& src);
    auto out_images_pop_front(cv::Mat& res) -> bool;
    auto out_images_empty() -> bool;

    void set_start_flag(bool value);
    auto get_start_flag() -> bool;

    void set_end_of_pre_processing(bool value);
    auto get_end_of_pre_processing() -> bool;

    void set_end_of_processing(bool value);
    auto get_end_of_processing() -> bool;

    void set_end_of_post_processing(bool value);
    auto get_end_of_post_processing() -> bool;

    auto get_end_of_stream() -> bool;
    void set_end_of_stream(bool value);

    void get_capture(cv::VideoCapture& capture);
};

#endif // GOLDWORKER_H
